; Programa: Package para entorno para agente de juego de los bloques
;           ver apuntes Tema 3 (ej. bsqueda en escalada: 
;           http://soma.etsii.ull.es/iia/transp/rp/img27.gif)
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: 6/7a (sistemas de produccin)
; Profesor: Patricio Garca (patricio@csi.ull.es)
; Fecha: 20/5/2004
; Comentarios:
;           Para representar los problemas se utilizan listas
;           tipo cols (ver (help 'cols)).
;           Las funciones accesibles mas importantes son:
;	       inic-bloq :   Comienza a ejecutar el agente
;              gen-bloq :    Devuelve lista con dos sublistas tipo COLS
;              vis-bloq :    Visualiza lista tipo cols
;              sol-val-bloq: Comprueba si se ha llegado al objetivo
;              mov-val-bloq: Determina si un movimiento es vido
;              mov-bloq:     Realiza un movimiento
;              version-bloq: Devuelve el nmero de versin
;
;           Ver el correspondiente help para cada una de ellas.
;
;           Para poder utilizarse este paquete ha de cargarse mediante:
;              (load "bloques")
;              (use-package "bloques")
;
; Modificaciones:
;           - Ver 0.5:
;               * Activacin de opcin :hof en inic-bloq
;           - Ver 0.4: 
;               * Posiblitar cargar check-bl-null.lsp en caso de
;                 que no exista check-bl.o y poder trabajar con otros
;                 lisp o versiones de gcl
;               * Solventar error de que col-max-bloq no devolva
;                 *col-max-bloq*, sino nil
;
;           - Ver 0.3: 
;               * Correccin de check-bl.o para que funcione sin
;                 el error de memory damage al utilizar :fich en inic-bloq.
;
; Por hacer:
;           - Probar bajo otros lisps
;           - Chekeos de consistencia de datos de entrada
;

(make-package "bloques" :nicknames '(blq bloq) :use '("LISP"))
(in-package "bloques")

(export '(version-bloques *verbose-bloq* col-max-bloq gen-bloq vis-bloq mov-val-bloq mov-bloq sol-val-bloq inic-bloq))


; Variables globales
; ------------------

(defvar *version-bloques* "0.4" "Versin de bloques.lsp")

(defun version-bloques ()
  " - Funcin: VERSION-BLOQUES ()
     Package:BLOQUES

Imprime la versin del paquete bloques que se ha cargado en memoria"
  *version-bloques*)

(defvar *verbose-bloq* nil 
  "- Variable: TIPO booleano

Activador de visualizacin traza de depuracin. T=activo, NIL=inactivo")


; Estructura de representacin del problema a resolver
; ----------------------------------------------------

(defvar col-max-def 10
  "- Variable: TIPO nmero 

Contiene el nmero columnas mximas por defecto en el juego de los bloques.")

(defvar *col-max-bloq* col-max-def
  "Nmero columnas mximas para el juego de los bloques actual.")

(defun col-max-bloq ()
  "- Funcin: COL-MAX-BLOQ ()
     Package:BLOQUES

Devuelve el nmero columnas mximas para el juego de los bloques actual."
  *col-max-bloq*)


; Funciones para creacion de la estructura de columnas
; ----------------------------------------------------

(defvar cols-ini-def '((A H G F E D C B))
  "Columnas iniciales por defecto")

(defvar cols-obj-def '((H G F E D C B A))
  "Columnas objetivo por defecto")

(defvar cols-ini-in cols-ini-def
  "Columnas iniciales entrada por usuario")

(defvar cols-obj-in cols-obj-def
  "Columnas objetivo entrada por usuario")

(defvar cols-ini cols-ini-def
  "Columnas iniciales")

(defvar cols-obj cols-obj-def
  "Columnas objetivo")

(defun filt-y-ord (cols)
  "Elimina las columnas vacas y las ordena en base al primer elemento de cada columna"
  (sort (remove-if 'null cols) 
        #'(lambda (e1 e2) 
	    (string-lessp (symbol-name (first (last e1))) 
			  (symbol-name (first (last e2)))))))


; Carga problema por defecto o desde fichero
; ------------------------------------------

(defun load-lab-def ()
  "Utiliza las columnas iniciales y objetivos por defecto"
  (setf cols-obj cols-obj-in)
  (setf cols-ini cols-ini-in)
  (list cols-ini cols-obj))

(defun load-lab-bloq (infile)
  "Lee pares de listas correspondientes a columnas iniciales y columnas objetivos"
  (setf cols-ini (filt-y-ord (read infile nil 'eof)))
  (setf cols-obj (filt-y-ord (read infile nil 'eof)))
  (list cols-ini cols-obj))

(defun load-lab-fich-bloq (fichero)
  ; Carga desde fichero
  (with-open-file (infile fichero)
		  (load-lab-bloq infile)))


; Generacion estructura inicial
; -----------------------------

(defvar fichero nil
  "Fichero de problema a cargar")

(defun gen-bloq ()
  " - Funcin: GEN-BLOQ ()
     Package:BLOQUES

Genera las listas de bloques iniciales y objetivos para el juego de Bloques, de acuerdo a las especificiaciones de los argumento INI, OBJ y FICH pasados a la funcin INIC-BLOQ (ver (help 'INIC-BLOQ). Devuelve una lista con dos elementos tipo COLS. El tipo COLS es una lista que contiene a su vez sublistas que representan las colmnas de bloqes, representados como smbolos, ordenadas por el elemento inferior y sin sublistas vacias (ver FILT-Y-ORD). El nmero mximo de columnas no puede superar el valor devuelto por COL-MAX-BLOQ, fijado por el parmetro COL-MAX de la fucin INIC-BLOQ."
  (cond
   ((null fichero) (load-lab-def))
   (t (load-lab-fich-bloq fichero))))


; Visualizacin de estructura de problema
; ---------------------------------------

(defun vis-bloq (cols)
  " - Funcin: VIS-BLOQ (COLS)
     Package:BLOQUES

Visualiza las columnas COLS. COLS ha de ser una lista tipo COLS, ver GEN-BLOQ."
  ; Tocar para que funcione bien con los smbolos de ms de 1 caracter
  (setf cols (cons nil cols))
  (format t "~%     ")
  (let ((max-alt (apply 'max (mapcar 'length cols))))
    (dotimes (i max-alt nil)
      (dolist (col cols nil)
	(if (>=  (- (+ i (length col)) max-alt) 0)
	    (format t " ~A " (symbol-name (nth (- (+ i (length col)) max-alt) col)))
	  (format t "   ")))
        (format t "~%     ")))
  (dolist (col cols nil)
    (format t "---"))
  (format t "~%#col:")
  (dotimes (i (length cols) nil)
    (format t " ~A " i))
  (terpri))


; Funciones para movimientos
; --------------------------

(defun mov-val-bloq (org des cols)
  " - Funcin: MOV-VAL-BLOQ (ORG DES COLS)
     Package:BLOQUES

Predicado que chequea si el movimiento de un bloque entre la columna origen ORG y la destino DES en la lista de columnas COLS es legal. COLS ha de ser una lista tipo COLS (ver GEN-BLOQ), ORG ha de ser un nmero mayor o igual a 1 y menor o igual al nmero de columnas en COLS, DES ha de ser un nmero entre 0 y el nmero de columnas existentes en COLS, si es 0 se supone que se mueve el bloque a la superficie y solo ser legal si no se ha llegado ya al mmero mximo de columnas permitidas en listas tipo COLS (ver COL-MAX-BLOQ). Devuelve T si es vlido el movimiento y NIL si no lo es."
  (and (>= org 1) (<= org (length cols)) ; Origen vlido
       (>= des 0) (<= des (length cols)) ; Destino vlido
       (/= org des)                      ; Origen distinto de destino
       (or (/= des 0) 
	   (and (< (length cols) *col-max-bloq*)
		(rest (nth (1- org) cols))))))
	   ; No se intenta sobrepasar en nmero mximo de columnas
           ; No se intenta mover bloque de columna con 1 a superficie

(defun punt-col (num cols)
  "Busca el puntero a una columna num (inicio en 0) dentro de cols"
  (dotimes (i num cols)
    (setf cols (rest cols))))

(defun mov-bloq (org des cols)
  " - Funcin: MOV-BLOQ (ORG DES COLS)
     Package:BLOQUES

Predicado realiza el movimiento de un bloque entre la columna origen ORG y la destino DES de la lista de columnas COLS. COLS ha de ser una lista tipo COLS (ver GEN-BLOQ), ORG ha de ser un nmero mayor o igual a 1 y menor o igual al nmero de columnas en COLS, DES ha de ser un nmero entre 0 y el nmero de columnas en COLS, si es 0 se supone que se mueve el bloque a la superficie y solo ser legal si no se ha llegado ya al mmero mximo de columnas permitidas en listas tipo COLS (ver COL-MAX-BLOQ). Devuelve una lista de columnas tipo COLS resultante del movimiento realizado."
  (let* ((nue-cols (cons nil (copy-alist cols))) ; Aade columna vaca
	 (punt-org (punt-col org nue-cols)) ; Puntero a columna org
	 (pblq (first punt-org)))           ; Puntero a primer bloque de org
    (setf (first punt-org) (rest (first punt-org)))   ; Elimino primer bloque de org
    (setf (rest pblq) (first (punt-col des nue-cols))) ; Bloque a mover apunta a primber bloque de dest
    (setf (first (punt-col des nue-cols)) pblq)       ; Introduzco bloque en dest
    (filt-y-ord nue-cols))) ; Filtra y ordena nuevas columnas


; Chequea si una solucin es vlida
;----------------------------------

(defun sol-val-bloq (cols)
  " - Funcin: SOL-VAL-BLOQ (COLS)
     Package:BLOQUES

Chequea si el las lista COLS es nula o constituye una lista objetivo. COLS ha de ser una lista tipo COLS (ver GEN-BLOQ)."
  (if cols
      (equal cols cols-obj)
    t))


; Cargamos Chequeador de soluciones
; ---------------------------------

(if (directory "check-bl.o")
    (load "check-bl")       ; Disponible en fichero check-bl.o
    (load "check-bl-null")) ; Sino se carga el suplente


; Lanzador de agentes y problemas
; -------------------------------

(defun getlistpartidas (fich)
  "Obtiene lista de partidas a partir de fich"
  (cond
   ((stringp fich) (directory fich))
   ((null fich) '(()))
   ((listp fich) (apply #'nconc (mapcar #'getlistpartidas fich)))
   (t '(()))))


; Comienza a ejecutar el agente pasado
; ------------------------------------
; Parmetros: 
;
; Variables globales afectadas:
;             verbose-bloq
;             fichero
;             cols-ini-in
;             cols-obj-in
;             cols-ini-def
;             cols-obj-def
;             col-max-def
;             *col-max-bloq*
;
; Valor de retorno:
;             (XITOS LONG-SOLS TIEMPO-EJEC)  
;                 Todo funcion correctamente
;             NIL Hubo algn problema
;
; Requerimientos:
;           * Exista una funcion AGENTE que encuentre una solucin
;             vlida al problema y devuelva los pasos de la solucin
;           * Si se instancia FICH, que sea un string empareje el 
;             nombre de fichero(s) existente(s) que contenga(n) 
;             las columnas iniciales y objetivos tipo COLS
;
; Ejemplo llamada: (INIC-BUMP :tipo 2) -> (1 13 12)

(defun inic-bloq (&optional
			 (agente #'agente-humano) ;jugador
			 ; debug        ; modo debug
			 &key 
			 verb         ; modo verbose
			 (num nil)    ; Nmero de partidas a jugar
			 (ini cols-ini-def)   ; Columnas iniciales
			 (obj cols-obj-def)   ; Columnas objetivos
			 (fich nil)   ; Fichero/s de partidas
			 (col-max col-max-def)    ; Nmero de columnas mximas
			 hof)         ; Aadir resultado al H.O.F.
  
  " - Funcin: INIC-BUMP (&OPTIONAL (AGENTE #'agente-humano) &KEY VERB (NUM 1) (INI COLS-INI-DEF) (OBJ COLS-OBJ-DEF) (FICH NIL) (COL-MAX COL-MAX-DEF) HOF)
     Package:BLOQUES

Inicia la ejecucion del agente AGENTE que juega a los Bloques. La funcion AGENTE debe encontrar una solucin y devolverla como una lista de movimientos, cada movimiento puede ser as su vez una lista con dos elementos: nmero de columna origen del movimiento y nmero de columna destino (ver MOV-VAL-BLOQ Y MOV-BLOQ). VERB indica si se activa el modo verbose. NUM indica el nmero de partidas a jugar. INI ha de contener una lista tipo COLS con las columnas iniciales del juego (ver GEN-BLOQ). OBJ ha de contener una lista tipo COLS con las columnas objetivos del juego. FICH indica el problema del juego a resolver, puede ser NIL (problema por defecto), o bien un string que cargar el laberinto desde el/los fichero(s) que empareje. Dichos ficheros estan compuesto de dos listas tipo COLS consecutivos. COL-MAX indica el nmero mximo de columnas que puden tener las listas tipo COLS. HOF indica si se graban los resultados en el Hall-of-Fame (/scratch/bloq/hof.txt , ver por*.txt).
Devuelve una lista con el nmero de partidas solucionadas con xito, el nmero acumulado de pasos de las soluciones encontradas, y el tiempo de ejecucion total que han tardado las partidas."

  (setf *verbose-bloq* verb)
  (setf cols-ini-in ini)
  (setf cols-obj-in obj)
  (setf fichero fich)
  (setf *col-max-bloq* col-max)
  (let (tiempo-ejec
	(listpart (getlistpartidas fich))
	listparact
	sol
	long-sol
	(long-sols 0)
	(fallos 0)
	(exitos 0))
    (if (null num)
	(setf num (length listpart)))
    (setf tiempo-ejec (get-internal-run-time))
    (dotimes (partida num)
      (if (null listparact)
	  (setf listparact listpart))
      (setf fichero (pop listparact))
      (setf sol (funcall agente))  ;Llamada al agente
      (setf long-sol (check-sol sol))
      (if (null long-sol)
	  (progn
	    (setf fallos (1+ fallos))
	    (format t "~%El agente fallo!!")))
      (if *verbose-bloq*
	  (progn
	    (format t "~%Partida:          ~D~%" (1+ partida))
	    (format t "Long. solucion:   ~D~%" long-sol)
	    (format t "Solucion:         ~s~%" sol)))
      (if long-sol
	  (setf long-sols (+ long-sols long-sol))))
    (setf tiempo-ejec (- (get-internal-run-time) tiempo-ejec))
    (setf exitos (- num fallos))
    (if *verbose-bloq*
	(progn
	  (format t "~%Nmero de partidas:            ~D~%" num)
	  (format t "Nmero de xitos:              ~D~%" exitos)
	  (format t "Tipo de partidas:              ~D~%" (unless fich (list ini obj)))
	  (format t "Tipo de problemas:             ~D~%" col-max)
	  (if (plusp exitos)
	      (progn
		(format t "Promedio Long. sol.:           ~D~%" (float (/ long-sols exitos)))
		(format t "Tiempo promedio por partida:   ~D~%~%" (float (/ tiempo-ejec exitos)))))))
    (if (and hof (plusp exitos))
      (printhof (gen-tipo fich col-max)
		agente 
		long-sols
		tiempo-ejec 
		exitos))
    (list exitos long-sols tiempo-ejec)))


